#include "../include/HPModelParser.h"
#include <cctype>
#include <exception>

bool parseHPModelJson(const std::string& jsonInput,
                      std::map<std::tuple<int,int,int>, Sphere>& spheres,
                      std::string& errorMessage)
{
    // Remove all whitespace
    std::string input;
    for (char c : jsonInput)
    {
        if (!std::isspace(static_cast<unsigned char>(c)))
            input += c;
    }

    if (input.empty())
    {
        errorMessage = "Input is empty.";
        return false;
    }

    if (input.front() != '[' || input.back() != ']')
    {
        errorMessage = "JSON must start with '[' and end with ']'.";
        return false;
    }

    input = input.substr(1, input.size() - 2); // Remove '[' and ']'

    size_t pos = 0;
    while (pos < input.size())
    {
        if (input[pos] != '{')
        {
            errorMessage = "Expected '{' at position " + std::to_string(pos);
            return false;
        }

        size_t endPos = input.find('}', pos);
        if (endPos == std::string::npos)
        {
            errorMessage = "Expected '}' after position " + std::to_string(pos);
            return false;
        }

        std::string objStr = input.substr(pos + 1, endPos - pos - 1);

        int x = 0, y = 0, z = 0;
        bool hydrophobic = false;
        bool xSet = false, ySet = false, zSet = false, hydroSet = false;

        size_t objPos = 0;
        while (objPos < objStr.size())
        {
            if (objStr[objPos] != '"')
            {
                errorMessage = "Expected '\"' for key at position " + std::to_string(objPos);
                return false;
            }
            size_t keyEnd = objStr.find('"', objPos + 1);
            if (keyEnd == std::string::npos)
            {
                errorMessage = "Expected '\"' to end key at position " + std::to_string(objPos + 1);
                return false;
            }

            std::string key = objStr.substr(objPos + 1, keyEnd - objPos - 1);

            size_t colonPos = objStr.find(':', keyEnd + 1);
            if (colonPos == std::string::npos)
            {
                errorMessage = "Expected ':' after key \"" + key + "\"";
                return false;
            }

            size_t valueStart = colonPos + 1;
            size_t valueEnd = valueStart;

            if (key == "hydrophobic")
            {
                if (objStr.compare(valueStart, 4, "true") == 0)
                {
                    hydrophobic = true;
                    valueEnd = valueStart + 4;
                }
                else if (objStr.compare(valueStart, 5, "false") == 0)
                {
                    hydrophobic = false;
                    valueEnd = valueStart + 5;
                }
                else
                {
                    errorMessage = "Invalid value for key \"hydrophobic\"";
                    return false;
                }
                hydroSet = true;
            }
            else
            {
                // Parse integer
                size_t numEnd = valueStart;
                if (objStr[numEnd] == '-') numEnd++;
                while (numEnd < objStr.size() && std::isdigit(static_cast<unsigned char>(objStr[numEnd])))
                {
                    numEnd++;
                }
                if (numEnd == valueStart)
                {
                    errorMessage = "Invalid integer value for key \"" + key + "\"";
                    return false;
                }
                std::string valueStr = objStr.substr(valueStart, numEnd - valueStart);
                int value = 0;
                try
                {
                    value = std::stoi(valueStr);
                }
                catch (std::exception&)
                {
                    errorMessage = "Invalid integer value for key \"" + key + "\"";
                    return false;
                }
                if (key == "x")
                {
                    x = value;
                    xSet = true;
                }
                else if (key == "y")
                {
                    y = value;
                    ySet = true;
                }
                else if (key == "z")
                {
                    z = value;
                    zSet = true;
                }
                else
                {
                    errorMessage = "Unexpected key \"" + key + "\"";
                    return false;
                }
                valueEnd = numEnd;
            }

            objPos = valueEnd;
            if (objPos < objStr.size() && objStr[objPos] == ',')
            {
                objPos++;
            }
        }

        if (!xSet || !ySet || !zSet || !hydroSet)
        {
            errorMessage = "Missing key(s) in object.";
            return false;
        }

        // Add sphere
        Sphere sphere;
        sphere.position = glm::vec3(x, y, -z);
        sphere.color = hydrophobic ? glm::vec3(1.0f,0.0f,0.0f) : glm::vec3(0.5f,0.5f,0.5f);
        spheres[std::make_tuple(x,y,z)] = sphere;

        pos = endPos + 1;
        if (pos < input.size() && input[pos] == ',')
        {
            pos++;
        }
    }

    return true;
}
