#include <iostream>
#include <vector>
#include <map>
#include <tuple>
#include <string>

// Include OpenGL / GLFW / GLAD
#include <glad/glad.h>
#include <GLFW/glfw3.h>

// glm includes (platform-specific)
#ifdef _WIN32
#include <glm.hpp>
#include <gtc/matrix_transform.hpp>
#include <gtc/type_ptr.hpp>
#else
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>
#endif

// ImGui includes
#include "imgui.h"
#include "imgui_impl_glfw.h"
#include "imgui_impl_opengl3.h"

// Our headers
#include "../include/Globals.h"
#include "../include/Input.h"
#include "../include/Spheres.h"
#include "../include/HPModelParser.h"
#include "../include/Rendering.h"
#include "../include/UI.h"

void framebuffer_size_callback(GLFWwindow* window, int width, int height)
{
    SCR_WIDTH = width;
    SCR_HEIGHT = height;
    glViewport(0, 0, SCR_WIDTH, SCR_HEIGHT);
}

int main()
{
    // Initialize GLFW
    if (!glfwInit())
    {
        std::cerr << "Failed to initialize GLFW" << std::endl;
        return -1;
    }

    // Set GLFW context version (OpenGL 3.3 Core)
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

    // Create window
    GLFWwindow* window = glfwCreateWindow(SCR_WIDTH, SCR_HEIGHT, "3D Spheres with Connections - UI Interaction Mode", NULL, NULL);
    if (window == NULL)
    {
        std::cout << "Failed to create GLFW window" << std::endl;
        glfwTerminate();
        return -1;
    }
    glfwMakeContextCurrent(window);

    // Set callbacks
    glfwSetFramebufferSizeCallback(window, framebuffer_size_callback);
    glfwSetCursorPosCallback(window, mouse_callback);
    glfwSetScrollCallback(window, scroll_callback);
    glfwSetKeyCallback(window, key_callback);

    // Initialize GLAD
    if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress))
    {
        std::cout << "Failed to initialize GLAD" << std::endl;
        return -1;
    }

    glEnable(GL_DEPTH_TEST);

    const char* vertexShaderSource = R"(
    #version 330 core
    layout (location = 0) in vec3 aPos;
    layout (location = 1) in vec3 aNormal;
    layout (location = 2) in vec3 aColor;

    out vec3 vertexColor;

    uniform vec3 sphereColor;
    uniform bool useUniformColor;

    uniform mat4 model;
    uniform mat4 view;
    uniform mat4 projection;

    void main()
    {
        if (useUniformColor)
            vertexColor = sphereColor;
        else
            vertexColor = aColor;
        gl_Position = projection * view * model * vec4(aPos, 1.0);
    }
    )";

    const char* fragmentShaderSource = R"(
    #version 330 core
    in vec3 vertexColor;
    out vec4 FragColor;

    void main()
    {
        FragColor = vec4(vertexColor, 1.0);
    }
    )";

    unsigned int shaderProgram = createShaderProgram(vertexShaderSource, fragmentShaderSource);
    if (shaderProgram == 0)
    {
        std::cerr << "Shader program creation failed!" << std::endl;
        return -1;
    }

    unsigned int sphereColorLoc = glGetUniformLocation(shaderProgram, "sphereColor");
    unsigned int useUniformColorLoc = glGetUniformLocation(shaderProgram, "useUniformColor");

    std::vector<float> sphereVertices;
    std::vector<unsigned int> sphereIndices;
    float sphereRadius = 0.1f;
    unsigned int sectorCount = 36;
    unsigned int stackCount = 18;
    generateSphere(sphereVertices, sphereIndices, sphereRadius, sectorCount, stackCount);

    unsigned int sphereVAO, sphereVBO, sphereEBO;
    glGenVertexArrays(1, &sphereVAO);
    glGenBuffers(1, &sphereVBO);
    glGenBuffers(1, &sphereEBO);

    glBindVertexArray(sphereVAO);
    glBindBuffer(GL_ARRAY_BUFFER, sphereVBO);
    glBufferData(GL_ARRAY_BUFFER, sphereVertices.size()*sizeof(float), &sphereVertices[0], GL_STATIC_DRAW);

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, sphereEBO);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sphereIndices.size()*sizeof(unsigned int), &sphereIndices[0], GL_STATIC_DRAW);

    // Position
    glVertexAttribPointer(0,3,GL_FLOAT,GL_FALSE,6*sizeof(float),(void*)0);
    glEnableVertexAttribArray(0);
    // Normal
    glVertexAttribPointer(1,3,GL_FLOAT,GL_FALSE,6*sizeof(float),(void*)(3*sizeof(float)));
    glEnableVertexAttribArray(1);
    glBindVertexArray(0);

    std::map<std::tuple<int,int,int>, Sphere> spheres;
    std::vector<float> lineVertices;
    unsigned int lineVAO, lineVBO;
    glGenVertexArrays(1,&lineVAO);
    glGenBuffers(1,&lineVBO);
    glBindVertexArray(lineVAO);
    glBindBuffer(GL_ARRAY_BUFFER, lineVBO);
    glBufferData(GL_ARRAY_BUFFER,0,nullptr,GL_DYNAMIC_DRAW);
    glVertexAttribPointer(0,3,GL_FLOAT,GL_FALSE,6*sizeof(float),(void*)0);
    glEnableVertexAttribArray(0);
    glVertexAttribPointer(2,3,GL_FLOAT,GL_FALSE,6*sizeof(float),(void*)(3*sizeof(float)));
    glEnableVertexAttribArray(2);
    glBindVertexArray(0);

    // Axes and grid setup
    const float AXIS_LENGTH = 10.0f;
    std::vector<float> axesVertices = {
        -AXIS_LENGTH,0.0f,0.0f, 1.0f,0.0f,0.0f,
         AXIS_LENGTH,0.0f,0.0f, 1.0f,0.0f,0.0f,
         0.0f,-AXIS_LENGTH,0.0f,0.0f,1.0f,0.0f,
         0.0f, AXIS_LENGTH,0.0f,0.0f,1.0f,0.0f,
         0.0f,0.0f, AXIS_LENGTH,0.0f,0.0f,1.0f,
         0.0f,0.0f,-AXIS_LENGTH,0.0f,0.0f,1.0f,
    };

    std::vector<float> gridPoints;
    int gridRange = (int)AXIS_LENGTH;
    for (int x=-gridRange; x<=gridRange; ++x)
    {
        for (int y=-gridRange; y<=gridRange; ++y)
        {
            for (int z=-gridRange; z<=gridRange; ++z)
            {
                gridPoints.push_back((float)x);
                gridPoints.push_back((float)y);
                gridPoints.push_back((float)z);
                gridPoints.push_back(1.0f);
                gridPoints.push_back(1.0f);
                gridPoints.push_back(1.0f);
            }
        }
    }

    unsigned int axesVAO, axesVBO;
    glGenVertexArrays(1,&axesVAO);
    glGenBuffers(1,&axesVBO);
    glBindVertexArray(axesVAO);
    glBindBuffer(GL_ARRAY_BUFFER, axesVBO);
    glBufferData(GL_ARRAY_BUFFER, axesVertices.size()*sizeof(float), axesVertices.data(), GL_STATIC_DRAW);
    glVertexAttribPointer(0,3,GL_FLOAT,GL_FALSE,6*sizeof(float),(void*)0);
    glEnableVertexAttribArray(0);
    glVertexAttribPointer(2,3,GL_FLOAT,GL_FALSE,6*sizeof(float),(void*)(3*sizeof(float)));
    glEnableVertexAttribArray(2);
    glBindVertexArray(0);

    unsigned int gridVAO, gridVBO;
    glGenVertexArrays(1,&gridVAO);
    glGenBuffers(1,&gridVBO);
    glBindVertexArray(gridVAO);
    glBindBuffer(GL_ARRAY_BUFFER, gridVBO);
    glBufferData(GL_ARRAY_BUFFER, gridPoints.size()*sizeof(float), gridPoints.data(), GL_STATIC_DRAW);
    glVertexAttribPointer(0,3,GL_FLOAT,GL_FALSE,6*sizeof(float),(void*)0);
    glEnableVertexAttribArray(0);
    glVertexAttribPointer(2,3,GL_FLOAT,GL_FALSE,6*sizeof(float),(void*)(3*sizeof(float)));
    glEnableVertexAttribArray(2);
    glBindVertexArray(0);

    // ImGui setup
    IMGUI_CHECKVERSION();
    ImGui::CreateContext();
    ImGuiIO& io = ImGui::GetIO();(void)io;
    ImGui::StyleColorsLight();
    ImGuiStyle& style = ImGui::GetStyle();
    style.WindowRounding = 0.0f;
    style.Colors[ImGuiCol_WindowBg] = ImVec4(1,1,1,1);

    ImGui_ImplGlfw_InitForOpenGL(window,true);
    ImGui_ImplOpenGL3_Init("#version 330");

    if (showUI)
        glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_NORMAL);
    else
        glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_DISABLED);

    while(!glfwWindowShouldClose(window))
    {
        float currentFrame = glfwGetTime();
        deltaTime = currentFrame - lastFrame;
        lastFrame = currentFrame;

        ImGui_ImplOpenGL3_NewFrame();
        ImGui_ImplGlfw_NewFrame();
        ImGui::NewFrame();

        if (showUI)
        {
            if (glfwGetInputMode(window, GLFW_CURSOR) != GLFW_CURSOR_NORMAL)
            {
                glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_NORMAL);
                firstMouse = true;
            }
        }
        else
        {
            if (glfwGetInputMode(window, GLFW_CURSOR) != GLFW_CURSOR_DISABLED)
            {
                glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_DISABLED);
                firstMouse = true;
            }
            processInput(window);
        }

        if (showUI)
        {
            renderUI(window, spheres, lineVertices, lineVBO);
        }

        glClearColor(0.1f,0.1f,0.1f,1.0f);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        glUseProgram(shaderProgram);
        glm::mat4 view = glm::lookAt(cameraPos, cameraPos+cameraFront, cameraUp);
        glm::mat4 projection = glm::perspective(glm::radians(fov),(float)SCR_WIDTH/(float)SCR_HEIGHT,0.1f,100.0f);

        unsigned int modelLoc = glGetUniformLocation(shaderProgram,"model");
        unsigned int viewLoc  = glGetUniformLocation(shaderProgram,"view");
        unsigned int projLoc  = glGetUniformLocation(shaderProgram,"projection");

        glUniformMatrix4fv(viewLoc,1,GL_FALSE,glm::value_ptr(view));
        glUniformMatrix4fv(projLoc,1,GL_FALSE,glm::value_ptr(projection));

        // Draw axis labels via ImGui overlay
        ImDrawList* drawList = ImGui::GetBackgroundDrawList();
        ImU32 textColor = IM_COL32(255,255,255,255);
        ImVec2 screenPos;
        if (WorldToScreen(glm::vec3(AXIS_LENGTH,0.0f,0.0f), view, projection, screenPos))
            drawList->AddText(screenPos,textColor,"X+");
        if (WorldToScreen(glm::vec3(-AXIS_LENGTH,0.0f,0.0f), view, projection, screenPos))
            drawList->AddText(screenPos,textColor,"X-");
        if (WorldToScreen(glm::vec3(0.0f,AXIS_LENGTH,0.0f), view, projection, screenPos))
            drawList->AddText(screenPos,textColor,"Y+");
        if (WorldToScreen(glm::vec3(0.0f,-AXIS_LENGTH,0.0f), view, projection, screenPos))
            drawList->AddText(screenPos,textColor,"Y-");
        if (WorldToScreen(glm::vec3(0.0f,0.0f,AXIS_LENGTH), view, projection, screenPos))
            drawList->AddText(screenPos,textColor,"Z-");
        if (WorldToScreen(glm::vec3(0.0f,0.0f,-AXIS_LENGTH), view, projection, screenPos))
            drawList->AddText(screenPos,textColor,"Z+");

        glUniform1i(useUniformColorLoc,0);
        // Draw axes
        glBindVertexArray(axesVAO);
        glUniformMatrix4fv(modelLoc,1,GL_FALSE,glm::value_ptr(glm::mat4(1.0f)));
        glDrawArrays(GL_LINES,0,6);
        glBindVertexArray(0);

        // Draw grid points
        if (showGridPoints)
        {
            glBindVertexArray(gridVAO);
            glPointSize(5.0f);
            glUniformMatrix4fv(modelLoc,1,GL_FALSE,glm::value_ptr(glm::mat4(1.0f)));
            glDrawArrays(GL_POINTS,0,(int)(gridPoints.size()/6));
            glBindVertexArray(0);
        }

        // Draw spheres
        glUniform1i(useUniformColorLoc,1);
        glBindVertexArray(sphereVAO);
        for (const auto& [key,sphere] : spheres)
        {
            glm::mat4 model = glm::mat4(1.0f);
            model = glm::translate(model,sphere.position);
            model = glm::scale(model, glm::vec3(sphereScale));
            glUniformMatrix4fv(modelLoc,1,GL_FALSE,glm::value_ptr(model));
            glUniform3f(sphereColorLoc, sphere.color.r, sphere.color.g, sphere.color.b);
            glDrawElements(GL_TRIANGLES, (int)sphereIndices.size(), GL_UNSIGNED_INT,0);
        }
        glBindVertexArray(0);

        // Draw lines
        glUniform1i(useUniformColorLoc,0);
        glLineWidth(lineThickness);
        glBindVertexArray(lineVAO);
        glUniformMatrix4fv(modelLoc,1,GL_FALSE,glm::value_ptr(glm::mat4(1.0f)));
        glDrawArrays(GL_LINES,0,(int)(lineVertices.size()/6));
        glBindVertexArray(0);
        glLineWidth(1.0f);

        ImGui::Render();
        ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());

        glfwSwapBuffers(window);
        glfwPollEvents();
    }

    glDeleteVertexArrays(1,&sphereVAO);
    glDeleteBuffers(1,&sphereVBO);
    glDeleteBuffers(1,&sphereEBO);

    glDeleteVertexArrays(1,&lineVAO);
    glDeleteBuffers(1,&lineVBO);

    glDeleteVertexArrays(1,&axesVAO);
    glDeleteBuffers(1,&axesVBO);

    glDeleteVertexArrays(1,&gridVAO);
    glDeleteBuffers(1,&gridVBO);

    glDeleteProgram(shaderProgram);

    ImGui_ImplOpenGL3_Shutdown();
    ImGui_ImplGlfw_Shutdown();
    ImGui::DestroyContext();

    glfwTerminate();
    return 0;
}
