//
// Created by Adam Wahab on 04/09/2024.
//

#include <fstream>
#include <iostream>
#include <sstream>
#include <string>
#include <filesystem>
#include "gtest/gtest.h"
#include "model/HPModel.h"
#include <stdexcept>

TEST(basic_check, test_eq) {
    EXPECT_EQ(1,1);
}

TEST(basic_check, test_fold_quantity) {
    HPModel model = HPModel("hp");
    model.fold(FORWARD);
    EXPECT_THROW(model.fold(FORWARD), std::invalid_argument);

    model = HPModel("pppp");
    model.fold(FORWARD);
    model.fold(FORWARD);
    model.fold(FORWARD);
    EXPECT_THROW(model.fold(FORWARD), std::invalid_argument);

    EXPECT_THROW(model = HPModel(""),  std::invalid_argument);
    EXPECT_THROW(model = HPModel("p"),  std::invalid_argument);
}

TEST(basic_check, test_circular) {
    HPModel model = HPModel("hpphp");
    model.fold(RIGHT);
    model.fold(RIGHT);
    model.fold(RIGHT);
    EXPECT_THROW(model.fold(RIGHT), std::invalid_argument);
}

TEST(basic_check, test_planar) {
    HPModel model = HPModel("ppppp");
    model.fold(FORWARD);
    model.fold(RIGHT);
    model.fold(FORWARD);
    model.fold(LEFT);
    auto points = model.getPoints();

    // Initial point
    Acid p0 = points[0];
    Acid p0Match = Acid(0, 0, 0, false);

    // Forward
    Acid p1 = points[1];
    Acid p1Match = Acid(0, 0, 1, false);

    // Right
    Acid p2 = points[2];
    Acid p2Match = Acid(1, 0, 1, false);

    // Forward
    Acid p3 = points[3];
    Acid p3Match = Acid(2, 0, 1, false);

    // Left
    Acid p4 = points[4];
    Acid p4Match = Acid(2, 0, 2, false);

    EXPECT_TRUE(p0.coords_match(p0Match));
    EXPECT_TRUE(p1.coords_match(p1Match));
    EXPECT_TRUE(p2.coords_match(p2Match));
    EXPECT_TRUE(p3.coords_match(p3Match));
    EXPECT_TRUE(p4.coords_match(p4Match));
}

TEST(basic_check, test_up_loop_valid) {
    HPModel model = HPModel("pppp");
    model.fold(UP);
    model.fold(UP);
    model.fold(UP);

    auto points = model.getPoints();

    // Initial Acid
    Acid p0 = points[0];
    Acid p0Match = Acid(0, 0, 0, false);

    // Up
    Acid p1 = points[1];
    Acid p1Match = Acid(0, 1, 0, false);

    // Up
    Acid p2 = points[2];
    Acid p2Match = Acid(0, 1, -1, false);

    // Up
    Acid p3 = points[3];
    Acid p3Match = Acid(0, 0, -1, false);

    EXPECT_TRUE(p0.coords_match(p0Match));
    EXPECT_TRUE(p1.coords_match(p1Match));
    EXPECT_TRUE(p2.coords_match(p2Match));
    EXPECT_TRUE(p3.coords_match(p3Match));
}

TEST(basic_check, test_up_relative_valid) {
    HPModel model = HPModel("ppppp");
    model.fold(UP);
    model.fold(RIGHT);
    model.fold(LEFT);
    model.fold(LEFT);

    auto points = model.getPoints();

    // Initial Acid
    Acid p0 = points[0];
    Acid p0Match = Acid(0, 0, 0, false);

    // Up
    Acid p1 = points[1];
    Acid p1Match = Acid(0, 1, 0, false);

    // Right
    Acid p2 = points[2];
    Acid p2Match = Acid(1, 1, 0, false);

    // Left
    Acid p3 = points[3];
    Acid p3Match = Acid(1, 2, 0, false);

    // Left
    Acid p4 = points[4];
    Acid p4Match = Acid(0, 2, 0, false);


    EXPECT_TRUE(p0.coords_match(p0Match));
    EXPECT_TRUE(p1.coords_match(p1Match));
    EXPECT_TRUE(p2.coords_match(p2Match));
    EXPECT_TRUE(p3.coords_match(p3Match));
    EXPECT_TRUE(p4.coords_match(p4Match));
}

TEST(basic_check, test_up_loop_invalid) {
    HPModel model = HPModel("ppppp");
    model.fold(UP);
    model.fold(UP);
    model.fold(UP);
    EXPECT_THROW(model.fold(UP), std::invalid_argument);
}


TEST(basic_check, test_up_down) {
    HPModel model = HPModel("pppppp");
    model.fold(RIGHT);
    model.fold(UP);
    model.fold(FORWARD);
    model.fold(DOWN);
    model.fold(LEFT);
    auto points = model.getPoints();

    // Initial Acid
    Acid p0 = points[0];
    Acid p0Match = Acid(0, 0, 0, false);

    // Right
    Acid p1 = points[1];
    Acid p1Match = Acid(1, 0, 0, false);

    // Up
    Acid p2 = points[2];
    Acid p2Match = Acid(1, 1, 0, false);

    // Forward
    Acid p3 = points[3];
    Acid p3Match = Acid(1, 2, 0, false);

    // Down
    Acid p4 = points[4];
    Acid p4Match = Acid(2, 2, 0, false);

    // Left
    Acid p5 = points[5];
    Acid p5Match = Acid(2, 2, 1, false);

    EXPECT_TRUE(p0.coords_match(p0Match));
    EXPECT_TRUE(p1.coords_match(p1Match));
    EXPECT_TRUE(p2.coords_match(p2Match));
    EXPECT_TRUE(p3.coords_match(p3Match));
    EXPECT_TRUE(p4.coords_match(p4Match));
    EXPECT_TRUE(p5.coords_match(p5Match));
}

TEST(basic_check, big_test) {
    HPModel model = HPModel("pppppppp");
    model.fold(DOWN);
    model.fold(DOWN);
    model.fold(RIGHT);
    model.fold(DOWN);
    model.fold(RIGHT);
    model.fold(UP);
    model.fold(LEFT);

    auto points = model.getPoints();


    // Initial point
    Acid p0 = points[0];
    Acid p0Match = Acid(0, 0, 0, false);

    // Down
    Acid p1 = points[1];
    Acid p1Match = Acid(0, -1, 0, false);

    // Down
    Acid p2 = points[2];
    Acid p2Match = Acid(0, -1, -1, false);

    // Right
    Acid p3 = points[3];
    Acid p3Match = Acid(1, -1, -1, false);

    // Down
    Acid p4 = points[4];
    Acid p4Match = Acid(1, 0, -1, false);

    // Right
    Acid p5 = points[5];
    Acid p5Match = Acid(1, 0, 0, false);

    // Up
    Acid p6 = points[6];
    Acid p6Match = Acid(2, 0, 0, false);

    // Left
    Acid p7 = points[7];
    Acid p7Match = Acid(2, 1, 0, false);

    EXPECT_TRUE(p0.coords_match(p0Match));
    EXPECT_TRUE(p1.coords_match(p1Match));
    EXPECT_TRUE(p2.coords_match(p2Match));
    EXPECT_TRUE(p3.coords_match(p3Match));
    EXPECT_TRUE(p4.coords_match(p4Match));
    EXPECT_TRUE(p5.coords_match(p5Match));
    EXPECT_TRUE(p6.coords_match(p6Match));
    EXPECT_TRUE(p7.coords_match(p7Match));
}

TEST(basic_check, right_up) {
    HPModel model = HPModel("ppppp");
    model.fold(RIGHT);
    model.fold(UP);
    model.fold(UP);
    model.fold(RIGHT);

    auto points = model.getPoints();

    // Initial
    Acid p0 = points[0];
    Acid p0Match = Acid(0, 0, 0, false);

    // Right
    Acid p1 = points[1];
    Acid p1Match = Acid(1, 0, 0, false);

    // Up
    Acid p2 = points[2];
    Acid p2Match = Acid(1, 1, 0, false);

    // Up
    Acid p3 = points[3];
    Acid p3Match = Acid(0, 1, 0, false);


    // Right
    Acid p4 = points[4];
    Acid p4Match = Acid(0, 1, -1, false);


    EXPECT_TRUE(p0.coords_match(p0Match));
    EXPECT_TRUE(p1.coords_match(p1Match));
    EXPECT_TRUE(p2.coords_match(p2Match));
    EXPECT_TRUE(p3.coords_match(p3Match));
    EXPECT_TRUE(p4.coords_match(p4Match));
}

// JSON tests
std::string readJsonFile(const std::string& filename) {
    std::ifstream file(filename);

    if (!file.is_open()) {
        std::cerr << "Unable to open file: " << filename << std::endl;
        return "";
    }

    std::stringstream buffer;
    buffer << file.rdbuf();
    return buffer.str();
}

TEST(basic_check, json_points_encodes_correctly_with_no_hydrophobics) {
    HPModel model = HPModel("ppppp");
    model.fold(RIGHT);
    model.fold(UP);
    model.fold(UP);
    model.fold(RIGHT);

    auto modelJsonData = model.pointsToJson();

    std::string filename =  std::string(TEST_FILES_DIR) + "/points_no_hydrophobes.json";
    std::string premadeJsonData = readJsonFile(filename);
    EXPECT_EQ(modelJsonData, premadeJsonData);
}

TEST(basic_check, check_acid_point) {
    Acid acid = Acid(1, 2, 3, true);
    Point expectedPoint = Point(1, 2, 3);
    Point actualPoint = acid.getPoint();
    EXPECT_EQ(expectedPoint, actualPoint);
}

TEST(basic_check, basic_score_test) {
    HPModel model = HPModel("hhhhh");
    model.fold(RIGHT);
    model.fold(RIGHT);
    model.fold(RIGHT);
    model.fold(FORWARD);
    EXPECT_EQ(model.getScore(), 1);
}

TEST(basic_check, check_score_pair_equality) {
    Point p1 = Point(0, 0, 0);
    Point p2 = Point(1, 1, 1);
    Point p3 = Point(2, 2, 2);

    ScorePair pair1 = ScorePair(p1, p2);
    ScorePair pair2 = ScorePair(p2, p1);
    ScorePair pair3 = ScorePair(p3, p1);

    EXPECT_EQ(pair1, pair2);
    EXPECT_EQ(pair2, pair1);
    EXPECT_TRUE(pair1 == pair2);
    EXPECT_FALSE(pair1 == pair3);
}

TEST(basic_check, test_model_clone) {
    HPModel model = HPModel("ppppp");

    model.fold(UP);
    model.fold(RIGHT);
    model.fold(LEFT);
    HPModel clone = model.clone();
    model.fold(LEFT);
    clone.fold(RIGHT);

    auto points = model.getPoints();
    auto clonePoints = clone.getPoints();

    // Initial Acid
    Acid p0 = points[0];
    Acid p0c = clonePoints[0];
    Acid p0Match = Acid(0, 0, 0, false);

    // Up
    Acid p1 = points[1];
    Acid p1c = clonePoints[1];
    Acid p1Match = Acid(0, 1, 0, false);

    // Right
    Acid p2 = points[2];
    Acid p2c = clonePoints[2];
    Acid p2Match = Acid(1, 1, 0, false);

    // Left
    Acid p3 = points[3];
    Acid p3c = clonePoints[3];
    Acid p3Match = Acid(1, 2, 0, false);

    // Left (Right for clone)
    Acid p4 = points[4];
    Acid p4c = clonePoints[4];
    Acid p4Match = Acid(0, 2, 0, false);
    Acid p4CloneMatch = Acid(2, 2, 0, false);


    // Check model is correct
    EXPECT_TRUE(p0.coords_match(p0Match));
    EXPECT_TRUE(p1.coords_match(p1Match));
    EXPECT_TRUE(p2.coords_match(p2Match));
    EXPECT_TRUE(p3.coords_match(p3Match));
    EXPECT_TRUE(p4.coords_match(p4Match));

    // Check clone is correct
    EXPECT_TRUE(p0c.coords_match(p0Match));
    EXPECT_TRUE(p1c.coords_match(p1Match));
    EXPECT_TRUE(p2c.coords_match(p2Match));
    EXPECT_TRUE(p3c.coords_match(p3Match));
    EXPECT_TRUE(p4c.coords_match(p4CloneMatch));
}

// TODO
TEST(basic_check, json_points_encodes_correctly_with_hydrophobics) {

}